#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "stream.h"
#include "globals.h"
#include "client.h"
#include "handle.h"
struct stream
{
  struct aisl_stream     _public;

  /* private data */
  list_t                 headers; /* request headers */
  buffer_t               buffer;

  const char            *c_type;

  aisl_http_response_t   response;
  stream_state_t         state;
  uint32_t               c_length;
  uint32_t               c_offset;
  int                    id;
  int                    flags;

  bool                   c_length_unknown;

};
static void
pair_free( pair_t self )
{
  if (!self) return;

  if(self->key) free(self->key);
  if(self->value) free(self->value);
  free(self);
}
/* -------------------------------------------------------------------------- */

stream_t
stream_new(struct sockaddr_in  *client, int id, stream_state_t state)
{
  stream_t self = malloc(sizeof(struct stream));

  if (self)
  {
    /* public data */
    ASTREAM(self)->client         = client;
    ASTREAM(self)->host           = NULL;
    ASTREAM(self)->path           = NULL;
    ASTREAM(self)->query          = NULL;
    ASTREAM(self)->scheme         = NULL;
    ASTREAM(self)->u_ptr          = NULL;
    ASTREAM(self)->request_method = AISL_HTTP_GET;

    /* private data */
    self->headers                 = NULL; /* request headers */
    self->buffer                  = buffer_new(0);

    self->c_type                  = NULL;

    self->response                = AISL_HTTP_OK;
    self->state                   = STREAM_REQUEST_METHOD;
    self->c_length                = 0;
    self->c_offset                = 0; /* headers length */
    self->id                      = id;
    self->c_length_unknown        = true;
    self->flags                   = 0;
  }
  return self;
}

/*
stream_t
stream_reset(stream_t self)
{
  if (ASTREAM(self)->path)
  {
    free( (char*) ASTREAM(self)->path);
    ASTREAM(self)->path = NULL;
  }

  if (ASTREAM(self)->query)
  {
    free( (char*) ASTREAM(self)->query);
    ASTREAM(self)->query = NULL;
  }

  ASTREAM(self)->u_ptr          = NULL;
  ASTREAM(self)->request_method = AISL_HTTP_GET;

  if (self->headers)
  {
    list_free(self->headers, (list_destructor_t) pair_free);
    self->headers = NULL;
  }

  self->c_type                  = NULL;
  self->response                = AISL_HTTP_OK;
  self->state                   = STREAM_REQUEST_METHOD;
  self->c_length                = 0;
  self->c_offset                = 0; / * headers length * /
  self->c_length_unknown        = true;
  self->flags                   = 0;

  return self;
}
*/
/* -------------------------------------------------------------------------- */

void
stream_free(stream_t self)
{
  if (self->buffer)  buffer_free(self->buffer);
  if (self->headers) list_free(self->headers, (list_destructor_t) pair_free);

  if (ASTREAM(self)->path)  free( (char*) ASTREAM(self)->path);
  if (ASTREAM(self)->query) free( (char*) ASTREAM(self)->query);

  aisl_handle_t hd =  ((client_t) ASTREAM(self)->client)->server->owner;
  aisl_raise_event(hd, self, AISL_STREAM_CLOSE);
  ASTREAM(self)->u_ptr = NULL;
  aisl_remove_listeners_for(hd, self);
  free(self);
}

/* -------------------------------------------------------------------------- */

int
stream_write(stream_t self, const char * data, uint32_t d_len)
{
  return buffer_add( self->buffer, data, d_len);
}

/* -------------------------------------------------------------------------- */

__attribute__ ((visibility ("default") ))
void
aisl_cancel(aisl_stream_t s)
{
  client_close( (client_t) s->client );
}

/* -------------------------------------------------------------------------- */

__attribute__ ((visibility ("default") ))
bool
aisl_is_secure(aisl_stream_t s)
{
  client_t cli = (client_t) s->client;

  return (cli->ssl) ? true : false;
}

/* -------------------------------------------------------------------------- */

__attribute__ ((visibility ("default") ))
void *
aisl_get_context(aisl_stream_t s)
{
  return s->u_ptr;
}

/* -------------------------------------------------------------------------- */

__attribute__ ((visibility ("default") ))
void
aisl_set_context(aisl_stream_t s, void * u_ptr)
{
  s->u_ptr = u_ptr;
}

/* -------------------------------------------------------------------------- */

__attribute__ ((visibility ("default") ))
aisl_client_t
aisl_get_client(aisl_stream_t s)
{
  return s->client;
}

/* -------------------------------------------------------------------------- */

__attribute__ ((visibility ("default") ))
aisl_server_t
aisl_get_server(aisl_stream_t s)
{
  return  (aisl_server_t) (((client_t) s->client)->server);
}


/* -------------------------------------------------------------------------- */

__attribute__ ((visibility ("default") ))
aisl_http_version_t
aisl_get_http_version(aisl_stream_t s)
{
  client_t cli = (client_t) s->client;

  return cli->protocol;
}

/* -------------------------------------------------------------------------- */

__attribute__ ((visibility ("default") ))
void
aisl_reject(aisl_stream_t s)
{
  client_t cli = (client_t) s->client;

  client_close( cli );
}

/* -------------------------------------------------------------------------- */



static char *
get_response_begin(stream_t stream)
{
  char * r;
  client_t cli = CLIENT(ASTREAM(stream)->client);

  r = str_printf(
        "%s %d %s\r\n"
        "Server: AISL\r\n"
        "Connection: %s\r\n\r\n",
        aisl_http_version_to_string(cli->protocol),
        stream->response,
        aisl_http_response_to_string(stream->response),
        ((cli->flags & CLIENT_FLAG_KEEPALIVE) ? "keep-alive" : "close")
      );

  return r;
}

__attribute__ ((visibility ("default") ))
aisl_status_t
aisl_response(aisl_stream_t stream, aisl_http_response_t  status_code,
                                    const char           *content_type,
                                    uint32_t              content_length)
{
  char  * pch;
  int     l;


  /* check if those headers were already sent */
  if (STREAM(stream)->state > STREAM_REQUEST_READY) return AISL_IDLE;

  STREAM(stream)->response = status_code;
  STREAM(stream)->c_type   = content_type;
  STREAM(stream)->c_length = content_length;

  if ( !(pch = get_response_begin(STREAM(stream))) )
    return AISL_MALLOC_ERROR;

  l = strlen(pch);
  STREAM(stream)->c_offset = l-2;

  buffer_clear(STREAM(stream)->buffer, content_length);

  l = buffer_add( STREAM(stream)->buffer, pch, l );
  free(pch);

  if (l == BUFFER_EOB) return AISL_MALLOC_ERROR;

  if (content_length)
  {
    if (!aisl_header_printf( stream, "Content-Length", "%u", content_length ))
      return AISL_MALLOC_ERROR;
  }

  if (content_type)
  {
    if (!aisl_header( stream, "Content-Type", content_type ))
      return AISL_MALLOC_ERROR;
  }

  STREAM(stream)->state = STREAM_RESPONSE_HEADER;

  return AISL_SUCCESS;
}


__attribute__ ((visibility ("default") ))
aisl_status_t
aisl_flush(aisl_stream_t stream)
{
  stream_t s = STREAM(stream); 
  if ( ! s->c_length )
  {
    s->c_length = s->buffer->size - s->c_offset - 2;
    if (!aisl_header_printf(stream, "Content-Length", "%u", s->c_length))
      return AISL_MALLOC_ERROR;
  }

  /*
  fprintf(stdout, "(%lu bytes)------->\n", STREAM(stream)->buffer->size);
  fwrite(STREAM(stream)->buffer->data, 1, STREAM(stream)->buffer->size, stdout);
  fprintf(stdout, "<------\n");
  */
  s->state = STREAM_RESPONSE_READY;

  s->flags |= STREAM_FLAG_OUTPUT_READY;

  return AISL_SUCCESS;
}


__attribute__ ((visibility ("default") ))
int
aisl_header(aisl_stream_t stream, const char *key, const char *value)
{
  int ret;
  char * pch;

  if ( (pch = str_printf("%s: %s\r\n", key, value)) != NULL )
  {
    ret = strlen(pch);
    if ( buffer_insert(
           STREAM(stream)->buffer,
           STREAM(stream)->c_offset,
           pch,
           ret
         ) == BUFFER_EOB )
    {
      ret = -1;
    }
    else
      STREAM(stream)->c_offset += ret;

    free(pch);
  }
  else
    ret = -1;

  return ret;
}


__attribute__ ((visibility ("default") ))
int
aisl_header_printf(aisl_stream_t stream, const char *key,
                                         const char *f_value, ...)
{
  int ret;

  va_list arg;
  va_start(arg, f_value);

  ret = aisl_header_vprintf(stream, key, f_value, arg);

  va_end(arg);

  return ret;
}


__attribute__ ((visibility ("default") ))
int
aisl_header_vprintf(aisl_stream_t stream, const char *key,
                                          const char *format,
                                          va_list     args)
{
  int ret;
  char * value;

  if ( (value = str_vprintf(format, args)) != NULL )
  {
    ret = aisl_header( stream, key, value );
    free(value);
  }
  else
    ret = -1;

  return ret;
}


__attribute__ ((visibility ("default") ))
int
aisl_printf(aisl_stream_t stream, const char *format, ...)
{
  va_list arg;
  va_start(arg, format);

  int result = aisl_vprintf(stream, format, arg);

  va_end(arg);

  /* No need to update length there, because vprintf do that
   * 
   * if (STREAM(stream)->c_length_unknown)
    STREAM(stream)->c_length += result;
    */


  return result;

}


__attribute__ ((visibility ("default") ))
int
aisl_vprintf(aisl_stream_t stream, const char *format, va_list args)
{
  int    result;
  char * r;

  if ( (r = str_vprintf(format, args)) != NULL)
  {
    result = strlen(r);
    if (buffer_add(STREAM(stream)->buffer, r, result) == BUFFER_EOB)
      result = -1;

    free(r);
  }
  else
    result = -1;

  return result;
}


__attribute__ ((visibility ("default") ))
int
aisl_write(aisl_stream_t stream, const char *data, int d_len)
{
  if (d_len < 0)
    d_len = strlen(data);

  if (buffer_add(STREAM(stream)->buffer, data, d_len) == BUFFER_EOB)
    d_len = -1;

  return d_len;
}


__attribute__ ((visibility ("default") ))
int
aisl_puts(const char *str, aisl_stream_t stream)
{
  return aisl_write( stream, str, strlen(str));
}


